package sk.fiit.jim.agent.models;

import java.util.HashMap;

import static sk.fiit.jim.agent.models.EnvironmentModel.*;
import static sk.fiit.jim.log.LogType.AGENT_MODEL;

import java.util.Map;
import java.util.Map.Entry;

import sk.fiit.jim.Settings;
import sk.fiit.jim.agent.AgentInfo;
import sk.fiit.jim.agent.Side;
import sk.fiit.jim.agent.moves.Joint;
import sk.fiit.jim.agent.parsing.ParsedData;
import sk.fiit.jim.agent.parsing.ParsedDataObserver;
import sk.fiit.jim.log.Log;
import sk.fiit.jim.math.Angles;
import sk.fiit.jim.math.Vector3D;

/**
 *  AgentModel.java
 *  
 *  Data structure encapsulating information about agent's current state,
 *  such as joint positions, global position and rotations against global
 *  axis. Calculations are delegated to {@link AgentPositionCalculator} and
 *  {@link AgentRotationCalculator}. Vector3D calculations from accelerometer
 *  are being used in ./scripts/plan/plan.rb to determine weather the agent
 *  is standing or laying on the ground.
 *
 *@Title	Jim
 *@author	marosurbanec
 *@author	Androids
 */
public class AgentModel implements ParsedDataObserver{
	private static AgentModel instance = new AgentModel();
	double rotationX = 0.0;
	double rotationY = 0.0;
	double rotationZ = 3.0*Math.PI / 2.0;
	Vector3D position = Vector3D.cartesian(0, 0, 0);
	Vector3D velocity = Vector3D.cartesian(0, 0, 0);
	Vector3D pureBodyAcceleration = Vector3D.cartesian(0, 0, 0);
	private Map<Joint, Double> jointAngles = new HashMap<Joint, Double>();
	ParsedData lastDataReceived = new ParsedData();
	
	public AgentModel(){
		for (Joint joint : Joint.values())
			jointAngles.put(joint, 0.0);
	}
	
	public static AgentModel getInstance(){
		return instance;
	}
	
	public static double getJointAngleOf(Joint joint){
		return instance.jointAngles.get(joint);
	}
	
	public Vector3D globalize(Vector3D relative) {
		return relative.rotateOverY(rotationY).rotateOverZ(rotationZ).rotateOverX(rotationX).add(position);
	}
	
	public Vector3D relativize(Vector3D global) {
		return global.subtract(position).rotateOverX(-rotationX).rotateOverZ(-rotationZ).rotateOverY(-rotationY);
	}

	//player and side identification added by Androids
	public void processNewServerMessage(ParsedData data){
		if (data.PLAYER_ID != null && data.PLAYER_ID != 0)
			AgentInfo.playerId = data.PLAYER_ID;
		if (data.OUR_SIDE_IS_LEFT != null)
			AgentInfo.side = data.OUR_SIDE_IS_LEFT ? Side.LEFT : Side.RIGHT;
				
		lastDataReceived = data;
		getPosition(data);
		updateJointPositions(data);
		adjustRotationsFor(data.gyroscope);
		updateRotations(data);
		updatePureBodyAcceleration(data);
		updatePosition(data);
	}

	private void updateJointPositions(ParsedData data){
		for (Entry<Joint, Double> record : data.agentsJoints.entrySet())
			this.jointAngles.put(record.getKey(), record.getValue());
	}
	
	/*
	 * Methods calculate vector moves around vectors and axis. The calculations
	 * are retrieved from theory on vector moves (google search on internet).
	 * Added by Androids.
	 */
	private Vector3D rotateVectorAroundVector(Vector3D initialVec, Vector3D axisV, double radAngle){
		Vector3D resultVec = null;
		double c = Math.cos(radAngle);
		double s = Math.sin(radAngle);
		double array[][] = new double[3][3];
		
		array[0][0]= (axisV.getX() * axisV.getX()) * (1.0f - c) + c;
		array[0][1]= (axisV.getY() * axisV.getX()) * (1.0f - c) + (axisV.getZ() * s);
		array[0][2]= (axisV.getZ() * axisV.getX()) * (1.0f - c) - (axisV.getY() * s);
		array[1][0]=(axisV.getX() * axisV.getY()) * (1.0f - c) - (axisV.getZ() * s);
		array[1][1]=(axisV.getY() * axisV.getY()) * (1.0f - c) + c;
		array[1][2]=(axisV.getZ() * axisV.getY()) * (1.0f - c) + (axisV.getX() * s);
		array[2][0]=(axisV.getX() * axisV.getZ()) * (1.0f - c) + (axisV.getY() * s);
		array[2][1]=(axisV.getY() * axisV.getZ()) * (1.0f - c) - (axisV.getX() * s);
		array[2][2]=(axisV.getZ() * axisV.getZ()) * (1.0f - c) + c;
		
		resultVec = vectorTimesMatrix(array, initialVec);
		return resultVec;
	}
	
	//added by Androids
	private Vector3D rotateVectorAroundZAxis(Vector3D initialVec, double angleV){
		Vector3D resultVec = null;
		double x = initialVec.getX() * Math.cos(angleV) - initialVec.getY() * Math.sin(angleV);
		double y = initialVec.getX() * Math.sin(angleV) + initialVec.getY() * Math.cos(angleV);
		
		resultVec = Vector3D.cartesian(x, y, initialVec.getZ());
		return resultVec;
	}

	//added by Androids
	private Vector3D countRotatedXAxis(double angleRotateZ, double angleRotateY){
		Vector3D x = Vector3D.cartesian(1, 0, 0);
		Vector3D y= Vector3D.cartesian(0, 1, 0);
		Vector3D firstRotateX = rotateVectorAroundZAxis(x,angleRotateZ);
		Vector3D firstRotateY = rotateVectorAroundZAxis(y,angleRotateZ);
		
		Vector3D resultVec = rotateVectorAroundVector(firstRotateX, firstRotateY, angleRotateY);
		return resultVec;
	}
	
	//added by Androids
	private Vector3D vectorTimesMatrix(double[][] matrix, Vector3D vector){
		double x = vector.getX()*matrix[0][0] + vector.getX()*matrix[0][1] + vector.getX()*matrix[0][2] ;
		double y = vector.getY()*matrix[1][0] + vector.getY()*matrix[1][1] + vector.getY()*matrix[1][2] ;
		double z = vector.getZ()*matrix[2][0] + vector.getZ()*matrix[2][1] + vector.getZ()*matrix[2][2] ;
		
		Vector3D resultVec = Vector3D.cartesian(x, y, z);
		return resultVec;
	}
	
	//added by Androids
	private double getAngleBetweenTwoVectors(Vector3D firstVec, Vector3D secondVec){
		double angleV;
		angleV = Math.acos(firstVec.getX() * secondVec.getX() + firstVec.getY() * secondVec.getY() + firstVec.getZ() * secondVec.getZ());
		return angleV;
	}
	
	private void updatePureBodyAcceleration(ParsedData data){
		Vector3D oldX = Vector3D.cartesian(1, 0, 0);
		Vector3D oldY = Vector3D.cartesian(0, 1, 0);
		Vector3D oldZ = Vector3D.cartesian(0, 0, 1);
		
		Vector3D rotatedX = countRotatedXAxis(rotationZ, rotationY);
		Vector3D rotatedY = Vector3D.cartesian(rotatedX.getY(), rotatedX.getX(), rotatedX.getZ());
		Vector3D rotatedZ = Vector3D.cartesian(rotatedY.getX(), rotatedY.getZ(), rotatedY.getY());
		
		if (data.accelerometer == null)
			return;
		
		pureBodyAcceleration = Vector3D.cartesian( data.accelerometer.getX() -  Math.cos ( getAngleBetweenTwoVectors(oldX, rotatedX) ) * Settings.getDouble("gravityAcceleration")  ,
				data.accelerometer.getY() -  Math.cos ( getAngleBetweenTwoVectors(oldY, rotatedY) )  * Settings.getDouble("gravityAcceleration") ,
				data.accelerometer.getZ() -  Math.cos ( getAngleBetweenTwoVectors(oldZ, rotatedZ) ) * Settings.getDouble("gravityAcceleration")  );
		
		Log.log(AGENT_MODEL, "Pure body acceleration: [%.5f,%.5f,%.5f]", pureBodyAcceleration.getX() ,pureBodyAcceleration.getY(),pureBodyAcceleration.getZ());
	}

	private void adjustRotationsFor(Vector3D gyroscope){
		if (gyroscope == null)
			return;
		
		rotationX += Math.toRadians(gyroscope.getX() * TIME_STEP);
		rotationX = Angles.normalize(rotationX);
		rotationY += Math.toRadians(gyroscope.getY() * TIME_STEP);
		rotationY = Angles.normalize(rotationY);
		rotationZ += Math.toRadians(gyroscope.getZ() * TIME_STEP);
		rotationZ = Angles.normalize(rotationZ);
		
		Log.log(AGENT_MODEL, "Rotation: [%.5f,%.5f,%.5f]", rotationX ,rotationY,rotationZ);
	}
	
	private void updateRotations(ParsedData data){
		new AgentRotationCalculator(this).updateRotations(data);
	}
	
	private void updatePosition(ParsedData data){
		new AgentPositionCalculator(this).updatePosition(data);
	}
	
	private void getPosition(ParsedData data){
		if (isStanding())
			System.out.println("standing");
		else if (isOnGround()) {
			System.out.println("on ground");
			if (isLyingOnBack())
				System.out.println("on back");
			else if (isLyingOnBelly())
				System.out.println("on belly");
		}
	}
	
	//---------------------QUERIES, GETTERS AND SETTERS-------------------
	public boolean isStanding(){
		return lastDataReceived.accelerometer.getZ() > 7.0;
	}
	
	public boolean isStanding(ParsedData data){
		return data.accelerometer.getZ() > 7.0;
	}
	
	public boolean isOnGround(){
		return Math.abs(lastDataReceived.accelerometer.getY()) > 9.3; 
	}
	
	public boolean isLyingOnBack(){
		return isOnGround() && (lastDataReceived.accelerometer.getY() > 9.3) ;
	}
	
	public boolean isLyingOnBelly(){
		return isOnGround() && !isLyingOnBack();
	}
	
	public double getRotationX(){
		return rotationX;
	}

	public double getRotationY(){
		return rotationY;
	}

	public double getRotationZ(){
		return rotationZ;
	}

	public Vector3D getPosition(){
		return position;
	}
	
	public void setPosition(Vector3D position){
		this.position = position;
	}

	public void setRotationX(double rotationX){
		this.rotationX = rotationX;
	}

	public void setRotationY(double rotationY){
		this.rotationY = rotationY;
	}

	public void setRotationZ(double rotationZ){
		this.rotationZ = rotationZ;
	}
}