package sk.fiit.jim.agent.moves;

import static sk.fiit.jim.log.LogType.LOW_SKILL;
import sk.fiit.jim.agent.communication.Communication;
import sk.fiit.jim.agent.models.EnvironmentModel;
import sk.fiit.jim.agent.skills.HighSkill;
import sk.fiit.jim.log.Log;
import sk.fiit.jim.log.LogType;

/**
 *  LowSkill.java
 *  
 *  Encapsulates a collection of PHASEs, implementing intelligent
 *  switching between those phases. An instance of LowSkill comes into
 *  play when HighSkill.pickLowSkill returns this instance as a result.
 *  
 *  LowSkill is populated by values contained in an XML. It must possess
 *  a name and an initial phase. Transition between phases is based on temporal
 *  events coming from the server. A low skill can be in one of two states:
 *  either it is finalizing or not. Finalization is triggered by
 *  {@link HighSkill}'s execute() method. It is triggered when this instance
 *  of LowSkill has reached the phase with isFinal flag set to true. If
 *  {@link HighSkill}'s pickLowSkill method chooses another LowSkill,
 *  a finalization of this LowSkill is executed. The objective of finalization
 *  is to get the player in basic stance.
 *  
 *  Every time a new phase is executed, a new target state is calculated by
 *  {@link JointPlacement}. The message sent to the server is generated by
 *  {@link JointPlacement} as well.
 *  
 *@Title	Jim
 *@author	marosurbanec
 *@author	Androids
 */
public class LowSkill{
	
	public String name;
	public String initialPhase;
	Phase activePhase;
	private double nextTimePhaseEnds = 0.0;
//	private boolean finalizing = false; //added by Androids - helped by JIM developers
	
	public LowSkill(){}
	public LowSkill(String name, String initialPhase){
		this.name = name;
		this.initialPhase = initialPhase;
	}
	
	public void executeFinalisation(){
		if (!activePhase.isFinal || !phaseEnded())
			return;
		
		setNewActivePhase(activePhase.finalizationPhase);
//		finalizing = true;
	}
	
	public void reset(){
		this.activePhase = null;
		this.nextTimePhaseEnds = 0.0;
//		finalizing = false;
	}
	
	public void carryOn(){
		if (activePhase == null)
			setNewActivePhase(initialPhase);
		if (phaseEnded() && activePhase.next != null)
			setNewActivePhase(activePhase.next);
		if (phaseEnded() && activePhase.next == null && activePhase.isFinal)
			executeFinalisation();
		if (phaseEnded() && activePhase.next == null && !activePhase.isFinal)
			setNewActivePhase(Phase.EMPTY_PHASE.name);
		
		Log.debug(LOW_SKILL, "Currently active phase: %s", activePhase.name);
		JointPlacement.generateMessage();
		Communication.getInstance().addToMessage(JointPlacement.lastMessage);
	}
	
	private void setNewActivePhase(String phaseName){
		if (shouldSetTrueFlag())
			SkipFlags.setTrue(activePhase.setFlagTrue);
		activePhase = Phases.get(phaseName);
		//TODO possible stack overflow if all the phases have the same skipIfFlag
		if (canSkipThisPhase()){
			setNewActivePhase(activePhase.next);
			return;
		}
		if (shouldSetFalseFlag())
			SkipFlags.setFalse(activePhase.setFlagFalse);
		
		nextTimePhaseEnds = EnvironmentModel.SIMULATION_TIME + activePhase.duration;
		Log.log(LogType.LOW_SKILL, "Starting phase %s", activePhase.name);
		JointPlacement.calculateNewTargetState(activePhase);
	}
	
	private boolean shouldSetTrueFlag(){
		return activePhase != null && activePhase.setFlagTrue != null;
	}	
	
	private boolean canSkipThisPhase(){
		return SkipFlags.isTrue(activePhase.skipIfFlag);
	}
	
	private boolean shouldSetFalseFlag(){
		return activePhase.setFlagFalse != null;
	}

	public boolean ended(){
		return phaseEnded() && activePhase != null && activePhase.isFinal;
	}

	private boolean phaseEnded(){
		return (EnvironmentModel.SIMULATION_TIME - nextTimePhaseEnds) > -0.01d;
	}
	
	//added by Androids
//	public boolean isFinalizing() {
//		return finalizing;
//	}
	
	@Override
	public String toString(){
		return name;
	}
}