package sk.fiit.jim.agent.skills;

import static sk.fiit.jim.log.LogType.HIGH_SKILL;
import sk.fiit.jim.agent.moves.LowSkill;
import sk.fiit.jim.log.Log;

/**
 *  HighSkill.java
 *  
 *  HighSkill represents an action - such as walking to a point, turning
 *  or standing up. HighSkill's responsibility is to achieve this goal or
 *  report a failure in achieving it. When subclassing a high skill, 2 methods
 *  need to be implemented:
 *  	1. checkProgress
 *  	2. pickLowSkill
 *  
 *  See their respective description for details.
 *
 *@Title	Jim
 *@author	marosurbanec
 *@author	Androids
 */
public abstract class HighSkill{
	
	protected LowSkill currentSkill;
	protected boolean finalizing = false;
	private boolean ended = false;
	
	public final void execute() throws Exception {
		Log.debug(HIGH_SKILL, "Currently executing: %s", currentSkill);
		if (currentSkill != null && !currentSkill.ended()){
			checkProgress();
			currentSkill.carryOn();
			return;
		}
		
		LowSkill newSkill = pickLowSkill();
		Log.debug(HIGH_SKILL, "Current skill: %s, new skill: %s, finalizing: %s", currentSkill, newSkill, Boolean.toString(finalizing));
		
		if (currentSkill != null && !finalizing && (newSkill == null || newSkill != currentSkill)){
			currentSkill.executeFinalisation();
			currentSkill.carryOn();
			finalizing = true;	//not sure weather this does anything.
		}
		
		if (finalizing && currentSkill.ended())
			finalizing = false;
		
		if (currentSkill != newSkill && newSkill != null){
			newSkill.reset();
			newSkill.carryOn();
		}
		
		/*
		 * The commented part should be uncommented if the agent should not repeat
		 * the moves, but do one skill per replan. The isFinalizing() method is situated
		 * in LowSkill.java class, and is being used to repeat cyclic moves, while finish
		 * the ones that are not cyclic in one replan turn. For the purpose of turnament
		 * RoboCup FIIT 2011, we hed to change this option. Androids.
		 */
		if (currentSkill == newSkill && currentSkill != null) {
//			if (currentSkill.isFinalizing())
//				currentSkill.reset();
			currentSkill.carryOn();
		}
		
		if (!finalizing)
			currentSkill = newSkill;
		
		ended = (currentSkill == null);
	}
	
	/**
	 * Return the next skill that needs to be executed
	 * in order to fulfill goals of this HighSkill.
	 * 
	 * If no skill is returned (returns null), a finalization
	 * of the last one is executed. If finalization has succesfully
	 * completed, this HighSkill's instance is abandoned and no longer
	 * used
	 * 
	 * @return {@link LowSkill} next skill that needs to be executed. Null if the goal is met
	 */
	public abstract LowSkill pickLowSkill();
	
	/**
	 * Throw an exception if robot is found to be in
	 * an unexpected state, e.g. fallen on ground during walking.
	 * Causes this skill execution to be immediately terminated.
	 * 
	 * @throws Exception any kind of exception will do (because of JRuby's hierarchy)
	 */
	public abstract void checkProgress() throws Exception;
	
	public boolean isEnded(){
		return ended;
	}
	
	public LowSkill getCurrentSkill(){
		return currentSkill;
	}
	
	public void setCurrentSkill(LowSkill skill) {
		this.currentSkill = skill;
	}

	@Override
	public String toString(){
		return "Current skill " + currentSkill;
	}
}